@@ -1,3 +1,9 @@
+import threading
+
+# Thread-local data
+_data = threading.local()
+
+
 def _enable_logging(f):
     """Enable logging of SQL statements when Flask is in use."""
 
@@ -45,6 +51,7 @@ class SQL(object):
         import sqlalchemy
         import sqlalchemy.orm
         import sqlite3
+        import threading
 
         # Require that file already exist for SQLite
         matches = re.search(r"^sqlite:///(.+)$", url)
@@ -54,8 +61,10 @@ class SQL(object):
             if not os.path.isfile(matches.group(1)):
                 raise RuntimeError("not a file: {}".format(matches.group(1)))
 
-        # Create engine, disabling SQLAlchemy's own autocommit mode, raising exception if back end's module not installed
-        self._engine = sqlalchemy.create_engine(url, **kwargs).execution_options(autocommit=False)
+        # Create engine, disabling SQLAlchemy's own autocommit mode raising exception if back end's module not installed;
+        # without isolation_level, PostgreSQL warns with "there is already a transaction in progress" for our own BEGIN and
+        # "there is no transaction in progress" for our own COMMIT
+        self._engine = sqlalchemy.create_engine(url, **kwargs).execution_options(autocommit=False, isolation_level="AUTOCOMMIT")
 
         # Get logger
         self._logger = logging.getLogger("cs50")
@@ -63,10 +72,6 @@ class SQL(object):
         # Listener for connections
         def connect(dbapi_connection, connection_record):
 
-            # Disable underlying API's own emitting of BEGIN and COMMIT so we can ourselves
-            # https://docs.sqlalchemy.org/en/13/dialects/sqlite.html#serializable-isolation-savepoints-transactional-ddl
-            dbapi_connection.isolation_level = None
-
             # Enable foreign key constraints
             if type(dbapi_connection) is sqlite3.Connection:  # If back end is sqlite
                 cursor = dbapi_connection.cursor()
@@ -83,7 +88,9 @@ class SQL(object):
         disabled = self._logger.disabled
         self._logger.disabled = True
         try:
-            self.execute("SELECT 1")
+            connection = self._engine.connect()
+            connection.execute("SELECT 1")
+            connection.close()
         except sqlalchemy.exc.OperationalError as e:
             e = RuntimeError(_parse_exception(e))
             e.__cause__ = None
@@ -97,9 +104,13 @@ class SQL(object):
 
     def _disconnect(self):
         """Close database connection."""
-        if hasattr(self, "_session"):
-            self._session.remove()
-            delattr(self, "_session")
+        if hasattr(_data, self._name()):
+            getattr(_data, self._name()).close()
+            delattr(_data, self._name())
+
+    def _name(self):
+        """Return object's hash as a str."""
+        return str(hash(self))
 
     @_enable_logging
     def execute(self, sql, *args, **kwargs):
@@ -268,43 +279,25 @@ class SQL(object):
         # Join tokens into statement
         statement = "".join([str(token) for token in tokens])
 
-        # Connect to database
+        # If no connection yet
+        if not hasattr(_data, self._name()):
+
+            # Connect to database
+            setattr(_data, self._name(), self._engine.connect())
+
+        # Use this connection
+        connection = getattr(_data, self._name())
+
+        # Disconnect if/when a Flask app is torn down
         try:
-
-            # Infer whether Flask is installed
             import flask
-
-            # Infer whether app is defined
             assert flask.current_app
-
-            # If no sessions for any databases yet
-            if not hasattr(flask.g, "_sessions"):
-                setattr(flask.g, "_sessions", {})
-            sessions = getattr(flask.g, "_sessions")
-
-            # If no session yet for this database
-            # https://flask.palletsprojects.com/en/1.1.x/appcontext/#storing-data
-            # https://stackoverflow.com/a/34010159
-            if self not in sessions:
-
-                # Connect to database
-                sessions[self] = sqlalchemy.orm.scoping.scoped_session(sqlalchemy.orm.sessionmaker(bind=self._engine))
-
-                # Remove session later
-                if _teardown_appcontext not in flask.current_app.teardown_appcontext_funcs:
-                    flask.current_app.teardown_appcontext(_teardown_appcontext)
-
-            # Use this session
-            session = sessions[self]
-
+            def teardown_appcontext(exception):
+                self._disconnect()
+            if teardown_appcontext not in flask.current_app.teardown_appcontext_funcs:
+                flask.current_app.teardown_appcontext(teardown_appcontext)
         except (ModuleNotFoundError, AssertionError):
-
-            # If no connection yet
-            if not hasattr(self, "_session"):
-                self._session  = sqlalchemy.orm.scoping.scoped_session(sqlalchemy.orm.sessionmaker(bind=self._engine))
-
-            # Use this session
-            session = self._session
+            pass
 
         # Catch SQLAlchemy warnings
         with warnings.catch_warnings():
@@ -324,10 +317,10 @@ class SQL(object):
 
                 # Execute statement
                 if self._autocommit:
-                    session.execute(sqlalchemy.text("BEGIN"))
-                result = session.execute(sqlalchemy.text(statement))
+                    connection.execute(sqlalchemy.text("BEGIN"))
+                result = connection.execute(sqlalchemy.text(statement))
                 if self._autocommit:
-                    session.execute(sqlalchemy.text("COMMIT"))
+                    connection.execute(sqlalchemy.text("COMMIT"))
 
                 # Check for end of transaction
                 if command in ["COMMIT", "ROLLBACK"]:
@@ -358,12 +351,30 @@ class SQL(object):
 
                 # If INSERT, return primary key value for a newly inserted row (or None if none)
                 elif command == "INSERT":
-                    if self._engine.url.get_backend_name() in ["postgres", "postgresql"]:
-                        try:
-                            result = session.execute("SELECT LASTVAL()")
-                            ret = result.first()[0]
-                        except sqlalchemy.exc.OperationalError:  # If lastval is not yet defined in this session
-                            ret = None
+
+                    # If PostgreSQL
+                    if self._engine.url.get_backend_name() == "postgresql":
+
+                        # Return LASTVAL() or NULL, avoiding
+                        # "(psycopg2.errors.ObjectNotInPrerequisiteState) lastval is not yet defined in this session",
+                        # a la https://stackoverflow.com/a/24186770/5156190;
+                        # cf. https://www.psycopg.org/docs/errors.html re 55000
+                        result = connection.execute("""
+                            CREATE OR REPLACE FUNCTION _LASTVAL()
+                            RETURNS integer LANGUAGE plpgsql
+                            AS $$
+                            BEGIN
+                                BEGIN
+                                    RETURN (SELECT LASTVAL());
+                                EXCEPTION
+                                    WHEN SQLSTATE '55000' THEN RETURN NULL;
+                                END;
+                            END $$;
+                            SELECT _LASTVAL();
+                        """)
+                        ret = result.first()[0]
+
+                    # If not PostgreSQL
                     else:
                         ret = result.lastrowid if result.rowcount == 1 else None
 
@@ -389,6 +400,8 @@ class SQL(object):
             # Return value
             else:
                 self._logger.debug(termcolor.colored(_statement, "green"))
+                if self._autocommit:  # Don't stay connected unnecessarily
+                    self._disconnect()
                 return ret
 
     def _escape(self, value):
@@ -536,10 +549,3 @@ def _parse_placeholder(token):
 
     # Invalid
     raise RuntimeError("{}: invalid placeholder".format(token.value))
-
-
-def _teardown_appcontext(exception=None):
-    """Closes context's database connection, if any."""
-    import flask
-    for session in flask.g.pop("_sessions", {}).values():
-        session.remove()
